from inspect import iscoroutine, signature, Parameter, isclass, getfullargspec
from typing import Optional
from marshmallow import ValidationError, fields, Schema
from aiohttp import web

from .directives import get_directive


def get_default_args(func):
    sig = signature(func)
    return {
        k: v.default
        for k, v in sig.parameters.items()
        if v.default is not Parameter.empty
    }


def get_arg(request, arg_name, defaults):
    if arg_name == "request":
        return request

    # get args from directives
    try:
        return get_directive(arg_name)(request)
    except KeyError:
        pass

    # get args from path
    try:
        return request.match_info[arg_name]
    except KeyError:
        pass

    # get args from query
    try:
        return request.rel_url.query[arg_name]
    except KeyError:
        pass

    try:
        return defaults[arg_name]
    except KeyError:
        raise fields.ValidationError("Required argument")


def cast_arg(arg, kind: Optional = None):
    # fields.Integer()
    if isinstance(kind, fields.Field):
        arg = kind.deserialize(arg)
    # arg: fields.Integer
    elif isclass(kind) and issubclass(kind, fields.Field):
        arg = kind().deserialize(arg)
    # arg: RequestSchema
    elif isclass(kind) and issubclass(kind, Schema):
        arg = kind(strict=True, many=False).dump(arg).data
    # RequestSchema()
    elif isinstance(kind, Schema):
        kind.strict = True
        arg = kind.dump(arg).data
    # int, string
    elif callable(kind):
        arg = kind(arg)

    return arg


async def get_kwargs(request: web.Request, handler):
    defaults = get_default_args(handler)
    arg_spec = getfullargspec(handler)
    kwargs = {}
    errors = {}
    for arg_name in arg_spec.args:
        try:
            arg = get_arg(request, arg_name, defaults)
            arg = await arg if iscoroutine(arg) else arg
            arg = cast_arg(arg, arg_spec.annotations.get(arg_name))
        except ValidationError as e:
            errors[arg_name] = e.messages
        else:
            if arg is not None:
                kwargs[arg_name] = arg
    return kwargs, errors



import pytest
from marshmallow import fields, Schema

from aiohug.arguments import cast_arg


class TestSchema(Schema):
    arg = fields.Integer()


@pytest.mark.parametrize(
    "arg,annotation,casted",
    (
        ("5", fields.Integer, 5),
        ("5", fields.Integer(), 5),
        ({"arg": "5"}, TestSchema, {"arg": 5}),
        ({"arg": "5"}, TestSchema(), {"arg": 5}),
        ("5", int, 5),
    ),
)
def test_field_class(arg, annotation, casted):
    assert cast_arg(arg, annotation) == casted



_available_directives = {}


def directive(fn):
    _available_directives[fn.__name__] = fn
    fn._directive = True
    return fn


def get_directive(fn):
    return _available_directives[fn]


def get_available_directives():
    return _available_directives


@directive
async def body(request):
    return await request.json()



from aiohttp import web
from marshmallow import fields, Schema

from aiohug import RouteTableDef


def create_app():
    app = web.Application()
    return app


async def test_not_valid_field(test_client):
    routes = RouteTableDef()

    @routes.get("/number/{number}/")
    async def return_number(number: fields.Int()):
        return {"number": number}

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get(f"/number/notanumber/")
    assert resp.status == 409
    assert await resp.json() == {
        "data": {"number": ["Not a valid integer."]},
        "status": "error",
    }


async def test_not_valid_schema(test_client):
    routes = RouteTableDef()

    class RequestSchema(Schema):
        a = fields.Int()
        b = fields.Int()

    @routes.get("/")
    async def with_body(body: RequestSchema):
        return body

    app = create_app()
    app.add_routes(routes)
    client = await test_client(app)
    resp = await client.get("/", json={"a": "5", "b": "c"})
    assert resp.status == 409
    assert await resp.json() == {
        "data": {"body": {"b": ["Not a valid integer."]}},
        "status": "error",
    }



from aiohttp import web

RESPONSE_STATUS = 0
RESPONSE_BODY = 1


def process_response(resp):
    # plain text
    if type(resp) is str:
        resp = web.Response(text=resp)

    # json
    elif type(resp) is dict or type(resp) is list:
        resp = web.json_response(data=resp)

    # status and body
    elif (
        type(resp) is tuple
        and type(resp[RESPONSE_STATUS]) is int
        and (type(resp[RESPONSE_BODY]) is dict or type(resp[RESPONSE_BODY]) is list)
    ):
        resp = web.json_response(status=resp[RESPONSE_STATUS], data=resp[RESPONSE_BODY])

    return resp



from aiohttp import web
from aiohttp.web_routedef import RouteDef

from .arguments import get_kwargs
from . import shortcuts


def _handle(handler):
    async def _handler(request):
        kwargs, errors = await get_kwargs(request, handler)

        if errors:
            return web.json_response(
                {"status": "error", "data": errors}, status=web.HTTPConflict.status_code
            )

        resp = await handler(**kwargs)
        resp = shortcuts.process_response(resp)
        return resp

    _handler._original_handler = handler
    return _handler


class RouteTableDef(web.RouteTableDef):
    def route(self, method, path, **kwargs):
        def inner(handler):
            handler = _handle(handler)
            self._items.append(RouteDef(method, path, handler, kwargs))
            return handler

        return inner



import json

import pytest
from aiohttp import web

from aiohug.shortcuts import process_response


def test_plaintext():
    text = "pong"
    resp = process_response(text)
    expected = web.Response(text=text)
    assert resp.text == expected.text
    assert resp.content_type == expected.content_type


@pytest.mark.parametrize("body", ({"message": "hello"}, [{"message": "hello"}]))
def test_json(body):
    response = process_response(body)
    expected_response = web.Response(
        text=json.dumps(body), content_type="application/json"
    )
    assert response.body == expected_response.body
    assert response.content_type == expected_response.content_type


@pytest.mark.parametrize("body", ({"message": "hello"}, [{"message": "hello"}]))
def test_status_and_body(body):
    status = 201
    response = process_response((status, body))
    expected_response = web.Response(
        text=json.dumps(body), content_type="application/json", status=status
    )
    assert response.status == expected_response.status
    assert response.body == expected_response.body



from aiohttp import web
from marshmallow import fields, Schema
from aiohug import RouteTableDef


def create_app():
    app = web.Application()
    return app


async def test_ping(test_client):
    routes = RouteTableDef()

    @routes.get("/ping/")
    async def hello():
        return "pong"

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get("/ping/")
    assert resp.status == 200
    text = await resp.text()
    assert "pong" in text


async def test_ping_with_request(test_client):
    routes = RouteTableDef()

    @routes.get("/ping/")
    async def ping(request):
        return "pong"

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get("/ping/")
    assert resp.status == 200
    text = await resp.text()
    assert "pong" in text


async def test_hello(test_client):
    routes = RouteTableDef()

    @routes.get("/hello/{name}/")
    async def hello(name: fields.String(), greeting: fields.String() = "Hello"):
        return {"msg": f"{greeting}, {name}"}

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get("/hello/Lucy/")
    assert resp.status == 200
    assert {"msg": "Hello, Lucy"} == await resp.json()


async def test_json_body(test_client):
    routes = RouteTableDef()

    @routes.get("/")
    async def with_body(body):
        return body

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    payload = {"msg": "Hello, Lucy"}
    resp = await client.get("/", json=payload)
    assert resp.status == 200
    assert await resp.json() == payload


async def test_json_body_with_schema_class(test_client):
    routes = RouteTableDef()

    class RequestSchema(Schema):
        count = fields.Int()

    @routes.get("/")
    async def with_body(body: RequestSchema):
        return body

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get("/", json={"count": "5"})
    assert resp.status == 200
    assert await resp.json() == {"count": 5}


async def test_json_body_with_schema_instance(test_client):
    routes = RouteTableDef()

    class RequestSchema(Schema):
        count = fields.Int()

    @routes.get("/")
    async def with_body(body: RequestSchema()):
        return body

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get("/", json={"count": "5"})
    assert resp.status == 200
    assert await resp.json() == {"count": 5}


# async def test_json_body_with_wrong_mime_type():
#     assert False
#
#     # todo: send no json mime type


async def test_cast(test_client):
    routes = RouteTableDef()

    @routes.get("/number/{number}/")
    async def return_number(number: fields.Int()):
        return {"number": number}

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    number = 5
    resp = await client.get(f"/number/{number}/")
    assert resp.status == 200
    assert await resp.json() == {"number": number}


async def test_cast_error(test_client):
    routes = RouteTableDef()

    @routes.get("/number/{number}/")
    async def return_number(number: fields.URL()):
        return {"number": number}

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    number = 5
    resp = await client.get(f"/number/{number}/")
    assert resp.status == 409
    assert await resp.json() == {
        "data": {"number": ["Not a valid URL."]},
        "status": "error",
    }


async def test_variable_not_provided_with_base_types(test_client):
    routes = RouteTableDef()

    @routes.get("/")
    async def return_number(number: int):
        return {"number": number}

    app = create_app()
    app.add_routes(routes)

    client = await test_client(app)
    resp = await client.get('/')
    assert resp.status == 409
    assert await resp.json() == {
        "data": {"number": ["Required argument"]},
        "status": "error",
    }





import asyncio

from aiohttp import web
import pytest

from aiohug.arguments import get_arg
from unittest.mock import MagicMock


def test_get_request():
    request = "message_that_should_be_the_same"
    assert get_arg(request, "request", {}) is request


async def test_directive():
    body = {"data": {}}
    request = MagicMock()
    request.json = asyncio.coroutine(MagicMock(return_value=body))
    assert await get_arg(request, "body", {}) == body


def test_arg_from_path():
    request = MagicMock()
    request.match_info = {"name": "Lucy"}
    assert get_arg(request, "name", {}) == "Lucy"


def test_arg_from_query():
    request = MagicMock()
    request.match_info = {}
    request.rel_url.query = {"name": "Lucy"}
    assert get_arg(request, "name", {}) == "Lucy"


def test_return_default():
    request = MagicMock()
    request.match_info = {}
    request.rel_url.query = {}

    assert get_arg(request, "name", {"name": 5}) == 5



from __future__ import unicode_literals
from setuptools import setup, find_packages
import os

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
PROJECT_INI = os.path.join(PROJECT_DIR, "project.ini")

config = ConfigParser()
config.read(PROJECT_INI)


def get_config(opt):
    return config.get("project", opt)


NAME = get_config("name")
DESCRIPTION = get_config("description")
URL = get_config("url")
AUTHOR = "nonamenix"
AUTHOR_EMAIL = "nonamenix@gmail.com"
README = "README.rst"
README_TXT = "README.txt"
LONG_DESCRIPTION = open(os.path.join(PROJECT_DIR, README)).read()

REQUIREMENTS_FILE = "requirements.txt"
REQUIREMENTS = open(os.path.join(PROJECT_DIR, REQUIREMENTS_FILE)).readlines()

VERSION = get_config("version")
DEV_VERSION = os.environ.get("DEV_VERSION")
if DEV_VERSION:
    VERSION = "{}.dev{}".format(VERSION, DEV_VERSION)
    config.set("project", "version", VERSION)
    with open(PROJECT_INI, "w") as f:
        config.write(f)

# create a README.txt file from .md
with open(README_TXT, "wb") as f:
    f.write(LONG_DESCRIPTION.encode())

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    include_package_data=True,
    packages=find_packages(),
    install_requires=REQUIREMENTS,
)

# delete README.txt
os.remove(README_TXT)
